import yaml

# make pyyaml don't include tag when dump. such as "!!python/object:api.project_yaml"
# see https://stackoverflow.com/a/48823424/2952665
def noop(self, *args, **kw):
    pass
yaml.emitter.Emitter.process_tag = noop



import yaml

class ProjectYaml():
    FileName = "luna.yaml"

    def __init__(
            self,
            yamlText: str
    ):

        o = yaml.load(
            yamlText,
            Loader=yaml.SafeLoader
        )


        ProjectYaml._shouldNotEmpty(o, [
            "version",
            "kind",
            "name",
            "evaluators",
            "scorer"
        ])

        self.version = o.get("version")
        self.kind = o.get("kind")
        self.name = o.get("name")
        self.description = o.get("description")
        self.media = o.get("media")
        self.modelBasePath = o.get("modelBasePath", "/models")

        self.evaluators = [self.Evaluator(e) for e in o.get("evaluators")]
        self.scorer = self.Scorer(o.get("scorer"))

        # check duplicate evaluator name
        evNames = [e.name for e in self.evaluators]
        evUniqueNames = set(evNames)
        if len(evNames) != len(evUniqueNames):
            raise ValueError("Duplicate evaluator name")

    def toYaml(self) -> str:
        return yaml.dump(self)

    @classmethod
    def _shouldNotEmpty(self, o, labels, path = ""):
        for l in labels:
            if o.get(l) == None or o.get(l) == "":
                raise ValueError("'{}{}' is missing in {}".format(path, l, ProjectYaml.FileName))

    class Evaluator:
        def __init__(
            self,
            o
        ):
            ProjectYaml._shouldNotEmpty(o, [
                "name",
                "image"
            ], "evaluators[].")

            self.name = o.get("name")
            self.image = o.get("image")
            self.command = o.get("command")
            self.models = o.get("models")

    class Scorer:
        def __init__(
            self,
            o
        ):
            ProjectYaml._shouldNotEmpty(o, [
                "image"
            ], "scorer.")

            self.image = o.get("image")
            self.command = o.get("command")
            self.modelPath = o.get("evalOutputFilter", "*")
            self.scoreFieldName = o.get("scoreFieldName", "score")
            self.sort = o.get("sort", "desc")



from git import Repo, InvalidGitRepositoryError
import glob
import sys, os
import logging
from datetime import datetime
import time
from git import Commit
from luna_ml.api.project_yaml import ProjectYaml
from luna_ml.api.model_yaml import ModelYaml

class RepoReader():
    def __init__(self, repoRootDir, diffOrigHash=None, diffCurrHash=None):
        self._repoRootDir = repoRootDir
        self._projects = {}

        self._repo = Repo(repoRootDir)

        if diffCurrHash == None: # working tree
            ts = datetime.utcfromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
            self._commit = f"working_tree {ts}"
        else:
            if isinstance(diffCurrHash, Commit):
                self._commit = diffCurrHash.hexsha
            else:
                self._commit = self._repo.rev_parse(diffCurrHash).hexsha

        # diff and scan
        diff = self._gitDiff(orig=diffOrigHash, current=diffCurrHash)
        self.scan(diff=diff)

    def getAll(self):
        return self._projects

    def scan(self, diff=None):
        lunaYamlPathList = []
        modelBasePathList = []

        # scan subdirs to find luna.yaml, and load
        for filename in glob.iglob("{}/**/{}".format(self._repoRootDir.rstrip("/"), ProjectYaml.FileName), recursive=True):
            lunaYamlPathList.append(filename)

        for filename in lunaYamlPathList:
            with open(filename, 'r') as file:
                yaml = file.read()
                projectYaml = ProjectYaml(yaml)

            projectBase = os.path.dirname(filename).rstrip("/")

            modelBasePath = "{}/{}".format(
                projectBase,
                projectYaml.modelBasePath.lstrip("/")
            )
            modelBasePathList.append(modelBasePath)

            # project resources
            resources = []
            allFilesUnderProject = glob.glob("{}/**".format(projectBase), recursive=True)
            allFilesUnderModelBase = glob.glob("{}/**".format(modelBasePath), recursive=True)
            notPartOfResource = [filename, modelBasePath]
            localEvalResult = glob.glob("{}/eval/**".format(projectBase), recursive=True)
            localScoreResult = glob.glob("{}/eval/**".format(projectBase), recursive=True)
            pycache = glob.glob("{}/__pycache__/**".format(projectBase), recursive=True)

            # filter files only
            allFilesUnderProject = [f for f in allFilesUnderProject if os.path.isfile(f)]

            # build project resources
            resources = list(set(allFilesUnderProject) - set(allFilesUnderModelBase) - set(notPartOfResource) - set(localEvalResult) - set(localScoreResult) - set(pycache))

            # retreive action
            if diff == None:
                action = "update" # when no diff is provided, consider it modified
                renamedFrom = None
            else:
                action, renamedFrom = self._getActionForProject(diff, os.path.dirname(filename), modelBasePath)

            # get model dict
            models = {}
            for modelDir in glob.glob("{}/*".format(modelBasePath)):
                if not os.path.isdir(modelDir):
                    continue

                m = self._scan_model(modelDir, diff=diff)
                models[modelDir] = m

            self._projects[projectBase] = {
                "yaml": projectYaml,
                "models": models,
                "modelBasePath": modelBasePath, # absolute path
                "resources": resources, # list of resource files. absolute path
                "action": action, # action need to be taken for this project. "update", "delete", "rename". None for no change
                "renamedFrom": renamedFrom, # previous luna.yaml file path, in case of action is "rename"
                "commit": self._commit
            }

    def _scan_model(self, path, diff=None):
        path = path.rstrip("/")
        modelYamlPath = "{}/{}".format(path, ModelYaml.FileName)

        # resources
        allFilesUnderModel = glob.glob("{}/**".format(path), recursive=True)
        notPartOfResource = [modelYamlPath]
        pycache = glob.glob("{}/__pycache__/**".format(path), recursive=True)

        # filter files only
        allFilesUnderModel = [f for f in allFilesUnderModel if os.path.isfile(f)]

        resources = list(set(allFilesUnderModel) - set(notPartOfResource) - set(pycache))

        # if ModelYaml.FileName does not exists, create a default one
        if os.path.isfile(modelYamlPath):
            with open(modelYamlPath, 'r') as file:
                yaml = file.read()
                modelYaml = ModelYaml(yaml)
        else:
            modelYaml = ModelYaml.default(os.path.basename(os.path.dirname(modelYamlPath)))

        action, renamedFrom = self._getActionForModel(diff, path)

        return {
            "yaml": modelYaml,
            "resources": resources,
            "action": action,
            "commit": self._commit
        }

    def _gitDiff(self, orig, current=None):
        '''
        current:
          None for working tree
          'HEAD' for head
        '''

        diff = {
            "added": [],
            "modified": [],
            "renamed": [],
            "deleted": []
        }

        untracked = []
        for f in self._repo.untracked_files:
            untracked.append("{}/{}".format(self._repoRootDir, f))

        if orig == None and current == None:
            diff["added"] = untracked
            return diff

        if orig == None:
            origIndex = self._repo.index
        elif isinstance(orig, Commit):
            origIndex = orig
        else:
            origIndex = self._repo.rev_parse(orig)

        if current == None:
            currentIndex = None
        elif isinstance(current, Commit):
            currentIndex = current
        else:
            currentIndex = self._repo.rev_parse(current)

        diffIndex = origIndex.diff(currentIndex)
        for diffItem in diffIndex.iter_change_type("A"):
            diff["added"].append("{}/{}".format(self._repoRootDir, diffItem.b_path))

        for diffItem in diffIndex.iter_change_type("M"):
            diff["modified"].append("{}/{}".format(self._repoRootDir, diffItem.b_path))

        for diffItem in diffIndex.iter_change_type("R"):
            diff["renamed"].append(
                ("{}/{}".format(self._repoRootDir, diffItem.a_path),
                "{}/{}".format(self._repoRootDir, diffItem.b_path))
            )

        for diffItem in diffIndex.iter_change_type("D"):
            diff["deleted"].append("{}/{}".format(self._repoRootDir, diffItem.b_path))

        if current == None:
            diff["added"].extend(untracked)

        return diff

    def _getActionForProject(self, diff, projectPath, modelbasePath):
        filteredDiff = {
            "added": self._filterChangedFileListForProject(diff["added"], projectPath, modelbasePath),
            "modified": self._filterChangedFileListForProject(diff["modified"], projectPath, modelbasePath),
            "deleted": self._filterChangedFileListForProject(diff["deleted"], projectPath, modelbasePath),
            "renamed": self._filterChangedFileListForProject(diff["renamed"], projectPath, modelbasePath)
        }

        yamlPath = "{}/{}".format(projectPath, ProjectYaml.FileName)

        # action is delete when luna.yaml is deleted
        if yamlPath in filteredDiff["deleted"]:
            return ("delete", None)
        
        # action is rename when luna.yaml is renamed
        for ren in filteredDiff["renamed"]:
            if ren[1] == yamlPath:
                return ("rename", ren[0])

        # action is update for any other changes
        if len(filteredDiff["added"]) + len(filteredDiff["modified"]) + len(filteredDiff["deleted"]) + len(filteredDiff["renamed"]) > 0:
            return ("update", None)

        # no action required
        return (None, None)

    def _getActionForModel(self, diff, modelPath):
        filteredDiff = {
            "added": self._filterChangedFileListForModel(diff["added"], modelPath),
            "modified": self._filterChangedFileListForModel(diff["modified"], modelPath),
            "deleted": self._filterChangedFileListForModel(diff["deleted"], modelPath),
            "renamed": self._filterChangedFileListForModel(diff["renamed"], modelPath)
        }

        yamlPath = "{}/{}".format(modelPath, ModelYaml.FileName)

        # for now, model delete, rename is not supported.

        # action is update for any changes
        if len(filteredDiff["added"]) + len(filteredDiff["modified"]) + len(filteredDiff["deleted"]) + len(filteredDiff["renamed"]) > 0:
            return ("update", None)

        # no action required
        return (None, None)


    def _filterChangedFileListForProject(self, changedFileList, projectPath, modelbasePath):
        changedFileInTheProject = [f for f in changedFileList if (isinstance(f, tuple) and f[1].startswith(projectPath)) or (isinstance(f, str) and f.startswith(projectPath))]
        changedFileInTheProjectExcludeModels = [f for f in changedFileInTheProject if not ((isinstance(f, tuple) and f[1].startswith(modelbasePath)) or (isinstance(f, str) and f.startswith(modelbasePath)))]
        return changedFileInTheProjectExcludeModels

    def _filterChangedFileListForModel(self, changedFileList, modelPath):
        changedFileListForModel = [f for f in changedFileList if (isinstance(f, tuple) and f[1].startswith(modelPath)) or (isinstance(f, str) and f.startswith(modelPath))]
        return changedFileListForModel



import yaml

class ModelYaml():
    FileName = "model.yaml"

    def __init__(
            self,
            yamlText: str
    ):

        o = yaml.load(
            yamlText,
            Loader=yaml.SafeLoader
        )

        ModelYaml._shouldNotEmpty(o, [
            "version",
            "kind",
            "name"
        ])

        self.version = o.get("version")
        self.kind = o.get("kind")
        self.name = o.get("name")

    def toYaml(self) -> str:
        return yaml.dump(self)

    @classmethod
    def default(cls, name: str):
        yaml = f"""
        version: v1
        kind: luna-ml/model
        name: {name}
        """
        return ModelYaml(yaml)

    @classmethod
    def _shouldNotEmpty(self, o, labels, path = ""):
        for l in labels:
            if o.get(l) == None or o.get(l) == "":
                raise ValueError("'{}{}' is missing in {}".format(path, l, ModelYaml.FileName))



#!/usr/bin/env python

import os, sys, io, time
import fire
import requests
import logging
import json
from colorama import init
from luna_ml.repository.repo_reader import RepoReader
from luna_ml.api.project_yaml import ProjectYaml
from luna_ml.api.model_yaml import ModelYaml

init()
from colorama import Fore, Back, Style

class LunaCli(object):
    """Luna ML CLI."""

    def __init__(self, server="http://localhost:5000"):
        self._server = server.rstrip("/")

    def __json_from_response(self, resp):
        return json.loads(resp.text)

    def __get_bool_param(self, boolParam, defaultValue: bool):
        if boolParam == None or boolParam == "":
            return defaultValue

        if boolParam == True or boolParam == False:
            return boolParam

        v = boolParam.strip().lower()
        if "true" == v:
            return True

        if "false" == v:
            return False

        return None

    def sync(self, local_repo, orig_commit=None, current_commit=None, dry_run=False, eval=True, wait=False):
        """Sync local git repository to update leaderboard

        sync command take changes of local git repository and update
        the server. By default, it takes changes between the HEAD and
        the curent working tree. Optionally, you can make take changes
        between two different commits by specifying flags.

        e.g.
            # sync changes of the local github repo in current directory
            luna-ml sync .

            # sync changes of the local github repo in absolute path
            luna-ml sync /my/repo
        """

        dry_run = self.__get_bool_param(dry_run, False)
        eval = self.__get_bool_param(eval, True)
        wait = self.__get_bool_param(wait, False)

        repoPath = os.path.abspath(local_repo).rstrip("/")
        print(f"Scanning changes from local repo ({repoPath}) ...")

        r = RepoReader(
                repoPath,
                diffOrigHash=orig_commit,
                diffCurrHash=current_commit
            )

        projects = r.getAll()

        if 0 == len(projects):
            print("No changes detected")
            return

        print("")
        self._printProjectSummaryAsTable(projects, repoPath)
        print("")

        if dry_run:
            return

        if not "REPO_TYPE" in os.environ:
            logging.error("REPO_TYPE environment is not defined. REPO_TYPE is something like 'github.com'")
            sys.exit(1)

        if not "REPO_NAME" in os.environ:
            logging.error("REPO_NAME environment is not defined. REPO_NAME is something like 'owner/repo'")
            sys.exit(1)

        repoType = os.environ["REPO_TYPE"]
        repoName = os.environ["REPO_NAME"]
        lunaAccessToken = os.environ["LUNA_ACCESS_TOKEN"]

        # post changes
        for projectPath, project in projects.items():
            # projectPath is absolute path
            lunaYamlPath = f"{projectPath}/{ProjectYaml.FileName}"
            projectPathRelativeToRepoRoot = projectPath[len(repoPath):]
            modelBasePath = "{}/{}".format(
                projectPath,
                project["yaml"].modelBasePath.lstrip("/").rstrip("/")
            )

            # update project
            if project["action"] == "update":
                print("Update project '{}' ... ".format(project["yaml"].name), end='')
                files = [("file", (ProjectYaml.FileName, open(lunaYamlPath, "rb")))]

                for r in project["resources"]:
                    files.append(("file", (r[len(projectPath):], open(r, "rb"))))

                resp = requests.post(
                    f"{self._server}/v1/project",
                    data={
                        "access_token": lunaAccessToken,
                        "repo_type": repoType,
                        "repo_name": repoName,
                        "path": projectPathRelativeToRepoRoot,
                        "commit": project["commit"]
                    },
                    files=files
                )

                if not resp.status_code == 200:
                    print("failed {}. {}".format(resp.status_code, resp.text))
                    logging.error("Update project {} failed. {}".format(project["yaml"].name, resp.text))
                    sys.exit(1)

                respProject = self.__json_from_response(resp)
                projectId = respProject["id"]
                print("done. id: {}".format(projectId))
            else:
                # get project
                resp = requests.get(
                    f"{self._server}/v1/project",
                    params={
                        "repo_type": repoType,
                        "repo_name": repoName,
                        "path": projectPathRelativeToRepoRoot
                    }
                )

                if not resp.status_code == 200:
                    logging.warn("Project {} is not created.".format(project["yaml"].name))
                    continue

                respProject = self.__json_from_response(resp)
                projectId = respProject["id"]

            # update models
            for modelPath, model in project["models"].items():
                # modelPath is absolute path
                modelPathRelativeToModelBase = modelPath[len(modelBasePath):]
                if model["action"] == "update":
                    print("Update model '{}' ... ".format(model["yaml"].name), end='')

                    modelYamlPath = f"{modelPath}/{ModelYaml.FileName}"

                    files = []

                    if os.path.isfile(modelYamlPath):
                        files.append(("file", (ModelYaml.FileName, open(modelYamlPath, "rb"))))

                    for r in model["resources"]:
                        files.append(("file", (r[len(modelPath.rstrip("/")):], open(r, "rb"))))

                    resp = requests.post(
                        f"{self._server}/v1/project/{projectId}/model",
                        data={
                            "access_token": lunaAccessToken,
                            "repo_type": repoType,
                            "repo_name": repoName,
                            "path": modelPathRelativeToModelBase,
                            "commit": project["commit"]
                        },
                        files=files
                    )

                    if not resp.status_code == 200:
                        print("failed {}. {}".format(resp.status_code, resp.text))
                        logging.error("Update model {} failed. {}".format(project["yaml"].name, resp.text))
                        sys.exit(1)
                    respModel = self.__json_from_response(resp)
                    modelId = respModel["id"]
                    print("done. id: {}".format(modelId))

                    if not eval:
                        continue

                    # trigger evaluation
                    print("trigger evaluation ... ")
                    for ev in project["yaml"].evaluators:
                        resp = requests.post(
                            f"{self._server}/v1/project/{projectId}/model/{modelId}/eval",
                            data={
                                "access_token": lunaAccessToken,
                                "name": ev.name
                            }
                        )

                        if not resp.status_code == 200:
                            print("Trigger evaluation failed. {} ".format(resp.status_code))
                            sys.exit(1)

                        if not wait:
                            continue

                        taskList = self.__json_from_response(resp)
                        if len(taskList) == 0:
                            continue

                        task = taskList[0] # because we're evaluating one by one here
                        taskId = task["id"]

                        print(f"Wait for the evaluation to finish ... ")
                        while "exit_code" not in task or task["exit_code"] == None:
                            time.sleep(5)

                            resp = requests.get(
                                f"{self._server}/v1/project/{projectId}/model/{modelId}/eval/{taskId}"
                            )

                            if not resp.status_code == 200:
                                print(f"Failed to get evaluation task {taskId}. {resp.status_code}")
                                sys.exit(1)

                            task = self.__json_from_response(resp)

                        if "exit_code" in task:
                            if task["exit_code"] != 0:
                                print(f"Task exited with {task['exit_code']} exit code.")
                                sys.exit(1)
                            else:
                                continue
                        else:
                            # should not reach here.
                            sys.exit(1)

    def _printProjectSummaryAsTable(self, projects, repoPath):
        projectList = list(projects.items())
        projectList.sort(key=lambda item: item[1]["yaml"].name)

        title_len = 30
        Col_Margin = "   "

        # print header
        print("{}{:<30}{}{:<10}{}{:<30}{}".format(
            Style.DIM,
            "Project",
            Col_Margin,
            "Change",
            Col_Margin,
            "Path",
            Style.RESET_ALL
        ))

        # print rows
        for path, project in projectList:
            title=project["yaml"].name
            title=title if len(title) <= title_len else f"{title[:title_len-2]}.."
            print("{:<30}{}{:<10}{}{:<30}".format(
                title,
                Col_Margin,
                project["action"] if project["action"] else "none",
                Col_Margin,
                path[len(repoPath):]
            ))

            # print models
            modelList = list(project["models"].items())
            modelList.sort(key=lambda item: item[1]["yaml"].name)

            model_count = 1
            for path, model in modelList:
                title=model["yaml"].name
                title=title if len(title) <= title_len else f"{title[:title_len-2]}.."

                if model_count == 1:
                    Model_Indent = "{}{:>15}{}".format(
                        Style.DIM,
                        "Models    ",
                        Style.RESET_ALL,
                    )
                else:
                    Model_Indent = "{:<15}".format("")

                print("{}{:<15}{}{:<10}{}{:<30}".format(
                    Model_Indent,
                    title,
                    Col_Margin,
                    model["action"] if model["action"] else "none",
                    Col_Margin,
                    path[len(repoPath):]
                ))

                model_count = model_count + 1



import unittest
from repository.repo_reader import RepoReader
import tempfile, os
import shutil
from git import Repo

class TestStorage(unittest.TestCase):
    def setUp(self):
        self._tmpdir = tempfile.mkdtemp()
        self._repo = Repo.init(self._tmpdir)
    
    def tearDown(self):
        shutil.rmtree(self._tmpdir)

    def _create_luna_yaml(self, path, name, yaml=None):
        yaml = yaml if yaml is not None else f"""
        name: {name}
        version: v1
        kind: luna-ml/project
        evaluators:
        - name: ev1
          image: python:3.7
        scorer:
          image: python:3.7
        """

        absolutePath = "{}{}".format(self._tmpdir, path)
        os.makedirs(os.path.dirname(absolutePath), exist_ok=True)
        with open(absolutePath, "w") as luna_yaml:
            print(yaml, file=luna_yaml)

    def _create_file(self, path, content):
        absolutePath = "{}{}".format(self._tmpdir, path)
        os.makedirs(os.path.dirname(absolutePath), exist_ok=True)
        with open(absolutePath, "w") as fileo:
            print(content, file=fileo)

    def test_when_project_exists_on_root__then_should_able_to_list(self):
        # given
        self._create_luna_yaml("/luna.yaml", "proj1")

        # when
        r = RepoReader(self._tmpdir)

        # then
        self.assertEqual(1, len(r._projects))
        self.assertEqual("proj1", r._projects["{}".format(self._tmpdir)]["yaml"].name)

    def test_when_projects_exist_on_dirs__then_should_able_to_list(self):
        # given
        self._create_luna_yaml("/proj1/luna.yaml", "proj1")
        self._create_luna_yaml("/proj2/luna.yaml", "proj2")

        # when
        r = RepoReader(self._tmpdir)

        # then
        self.assertEqual(2, len(r._projects))

    def test_when_invalid_luna_yaml__then_should_raise_exception(self):
        # given
        self._create_luna_yaml("/proj1/luna.yaml", "proj1") # valid
        self._create_luna_yaml("/proj2/luna.yaml", "proj2", yaml=
            """
            name: proj2
            version: v1
            kind: luna-ml/project
            scorer:
                image: python:3.7
            """
        ) # invalid. missing evaluator

        # when
        self.assertRaises(Exception, RepoReader, self._tmpdir)

    def test_when_projects_exist_on_dirs__then_should_able_to_list_resources(self):
        # given 2 projects
        self._create_luna_yaml("/proj1/luna.yaml", "proj1")
        self._create_luna_yaml("/proj2/luna.yaml", "proj2")

        # given file not belongs to any project
        self._create_file("/file1", "v1")

        # given files belongs to project
        self._create_file("/proj1/file12", "v12")
        self._create_file("/proj1/subdir/file13", "v13")

        # given file under the model path
        self._create_file("/proj1/models/file14", "v14")

        # given file belongs to another project
        self._create_file("/proj2/file25", "v25")

        # when
        r = RepoReader(self._tmpdir)

        # then
        self.assertEqual(2, len(r._projects))

        # then
        proj1 = r._projects["{}/proj1".format(self._tmpdir)]
        self.assertEqual(2, len(proj1["resources"]))
        self.assertTrue("{}/proj1/file12".format(self._tmpdir) in proj1["resources"])
        self.assertTrue("{}/proj1/subdir/file13".format(self._tmpdir) in proj1["resources"])

        proj2 = r._projects["{}/proj2".format(self._tmpdir)]
        self.assertEqual(1, len(proj2["resources"]))
        self.assertTrue("{}/proj2/file25".format(self._tmpdir) in proj2["resources"])


    def test_when_models_exist__then_should_able_to_list(self):
        # given 2 projects
        self._create_luna_yaml("/proj1/luna.yaml", "proj1")
        self._create_luna_yaml("/proj2/luna.yaml", "proj2")

        # given file under the model path
        self._create_file("/proj1/models/file11", "v1") # should be ignored, since it is not inside of individual model dir

        # given file under the model path
        self._create_file("/proj1/models/m1/file12", "v112")
        self._create_file("/proj1/models/m1/sudir/file13", "v113")
        self._create_file("/proj1/models/m2/file23", "v123")

        # given file under the model in another project
        self._create_file("/proj2/models/m3/file24", "v234")
        self._create_file("/proj2/models/m3/model.yaml", """
        version: v1
        kind: luna-ml/model
        name: "my model 3"
        """)

        # when
        r = RepoReader(self._tmpdir)

        # then two projects
        self.assertEqual(2, len(r._projects))

        # then num models
        proj1 = r._projects["{}/proj1".format(self._tmpdir)]
        proj2 = r._projects["{}/proj2".format(self._tmpdir)]
        self.assertEqual(2, len(proj1["models"]))
        self.assertEqual(1, len(proj2["models"]))

        # then model name is dir name when model.yaml is missing
        model1 = proj1["models"]["{}/proj1/models/m1".format(self._tmpdir)]
        model2 = proj1["models"]["{}/proj1/models/m2".format(self._tmpdir)]
        model3 = proj2["models"]["{}/proj2/models/m3".format(self._tmpdir)]
        self.assertEqual("m1", model1["yaml"].name)
        self.assertEqual("m2", model2["yaml"].name)

        # then model name from yaml when model.yaml exists
        self.assertEqual("my model 3", model3["yaml"].name)

        # then model resources
        self.assertTrue(2, len(model1["resources"]))
        self.assertTrue(1, len(model2["resources"]))
        self.assertTrue(1, len(model3["resources"]))

    def test_filterChangedFileListForProject(self):
        r = RepoReader(self._tmpdir)

        self.assertEqual(
            ["/proj1/luna.yaml", "/proj1/res1"],
            r._filterChangedFileListForProject(
                changedFileList=["/proj1/luna.yaml", "/proj1/res1", "/proj1/models/file1", "/proj2/luna.yaml"],
                projectPath="/proj1",
                modelbasePath="/proj1/models"
        ))

        self.assertEqual(
            ["/proj2/luna.yaml"],
            r._filterChangedFileListForProject(
                changedFileList=["/proj1/luna.yaml", "/proj1/res1", "/proj1/models/file1", "/proj2/luna.yaml"],
                projectPath="/proj2",
                modelbasePath="/proj2/models"
        ))

        self.assertEqual(
            [],
            r._filterChangedFileListForProject(
                changedFileList=["/proj1/luna.yaml", "/proj1/res1", "/proj1/models/file1", "/proj2/luna.yaml"],
                projectPath="/proj3",
                modelbasePath="/proj3/models"
        ))


    def test_filterChangedFileListForModel(self):
        r = RepoReader(self._tmpdir)

        self.assertEqual(
            ["/proj1/models/m1/file1"],
            r._filterChangedFileListForModel(
                changedFileList=["/proj1/luna.yaml", "/proj1/res1", "/proj1/models/file1", "/proj1/models/m1/file1", "/proj1/models/m2/file2"],
                modelPath="/proj1/models/m1"
        ))

    def test_when_diff_on_working_tree__then_should_return_both_untracked_added(self):
        # given
        self._create_file("/untracked", "v1")
        self._create_file("/added", "v12")
        self._create_file("/commited", "v13")

        # when
        r = RepoReader(self._tmpdir)
        r._repo.git.add(f"{self._tmpdir}/commited")
        r._repo.index.commit("initial commit")
        r._repo.git.add(f"{self._tmpdir}/added")

        diff = r._gitDiff(orig="HEAD", current=None)

        # then
        self.assertEqual(2, len(diff["added"]))
        self.assertTrue(f"{self._tmpdir}/added" in diff["added"])
        self.assertTrue(f"{self._tmpdir}/untracked" in diff["added"])

    def test_when_diff_against_repo_without_commit__then_should_return_all_untracked_as_added(self):
        # given
        self._create_file("/untracked1", "v1")
        self._create_file("/untracked2", "v12")

        # when
        r = RepoReader(self._tmpdir)
        diff = r._gitDiff(orig=None, current=None)

        # then
        self.assertEqual(2, len(diff["added"]))
        self.assertTrue(f"{self._tmpdir}/untracked1" in diff["added"])
        self.assertTrue(f"{self._tmpdir}/untracked2" in diff["added"])

    def test_when_diff_between_two_commit__then_should_return_all_changes(self):
        # given
        self._create_file("/unchanged", "v1")
        self._create_file("/modified", "v11")
        self._create_file("/added", "v12")
        self._create_file("/deleted", "v13")
        self._create_file("/renamed-a", "v14")
        self._create_file("/untracked", "v15")

        # when
        r = RepoReader(self._tmpdir)
        r._repo.git.add(f"{self._tmpdir}/unchanged")
        r._repo.git.add(f"{self._tmpdir}/modified")
        r._repo.git.add(f"{self._tmpdir}/deleted")
        r._repo.git.add(f"{self._tmpdir}/renamed-a")
        initialCommit = r._repo.index.commit("initial commit")

        self._create_file("/modified", "v11-1")
        r._repo.git.add(f"{self._tmpdir}/modified")
        r._repo.git.add(f"{self._tmpdir}/added")
        r._repo.index.remove([f"{self._tmpdir}/deleted"])
        r._repo.index.remove([f"{self._tmpdir}/renamed-a"])
        self._create_file("/renamed-b", "v14")
        r._repo.git.add(f"{self._tmpdir}/renamed-b")
        currentCommit = r._repo.index.commit("second commit")

        # then
        diff = r._gitDiff(orig=initialCommit.hexsha, current=currentCommit.hexsha)
        self.assertEqual([f"{self._tmpdir}/added"], diff["added"])
        self.assertEqual([f"{self._tmpdir}/modified"], diff["modified"])
        self.assertEqual([f"{self._tmpdir}/deleted"], diff["deleted"])
        self.assertEqual([(f"{self._tmpdir}/renamed-a", f"{self._tmpdir}/renamed-b")], diff["renamed"])

    def test_when_project_create_in_empty_repo__then_action_should_be_update(self):
        # given
        self._create_luna_yaml("/luna.yaml", "proj1")

        # when
        r = RepoReader(self._tmpdir)

        # then
        proj = r._projects["{}".format(self._tmpdir)]
        self.assertEqual("update", proj["action"])

    def test_when_project_has_no_change__then_action_should_be_none(self):
        # given
        self._create_luna_yaml("/luna.yaml", "proj1")
        self._repo.git.add(f"{self._tmpdir}")
        initialCommit = self._repo.index.commit("initial commit")

        # when
        r = RepoReader(self._tmpdir)

        # then
        proj = r._projects["{}".format(self._tmpdir)]
        self.assertEqual(None, proj["action"])

    def test_when_project_is_modified__then_action_should_be_update(self):
        # given
        self._create_luna_yaml("/luna.yaml", "proj1")
        self._create_file("/res1", "resource1")
        self._repo.git.add(f"{self._tmpdir}/luna.yaml")
        self._repo.git.add(f"{self._tmpdir}/res1")
        initialCommit = self._repo.index.commit("initial commit")
        self._repo.index.remove(f"{self._tmpdir}/res1")
        secondCommit = self._repo.index.commit("remove resource")
        os.remove(f"{self._tmpdir}/res1")

        # when
        r = RepoReader(self._tmpdir, initialCommit, secondCommit)

        # then
        proj = r._projects["{}".format(self._tmpdir)]
        self.assertEqual("update", proj["action"])

    def test_when_model_add__then_action_should_be_update(self):
        # given
        self._create_luna_yaml("/luna.yaml", "proj1")
        self._create_file("/models/m1/file", "")

        # when
        r = RepoReader(self._tmpdir)

        # then
        proj = r._projects["{}".format(self._tmpdir)]
        model = proj["models"]["{}/models/m1".format(self._tmpdir)]
        self.assertEqual("update", model["action"])
